import re
from abc import ABC, abstractmethod

from .utils import parse_hex, per_join


class Formatter(ABC):
    def _to_clean_str(self, value, max_value):
        value_str = str(hex(value))[2:]
        full_mac_str = '0' * (self._hex_digits_count(max_value) -
                              len(value_str)) + value_str
        return full_mac_str

    def _hex_digits_count(self, value):
        return len(hex(value)[2:])

    def _common_regex(self, max_value, delimeter, step):
        return "^" + per_join([r'[\da-fA-f]' for _ in range(self._hex_digits_count(max_value))], delimeter, step) + "$"

    @abstractmethod
    def format(self, value, max_value):
        pass

    def parse(self, _str, max_value):
        m = re.match(self._get_validator_regexp(_str, max_value), _str)
        if m is None:
            raise ValueError('Invalid MAC address format')
        else:
            return self._parse_value_from_str(_str)

    def _parse_value_from_str(self, _str):
        return parse_hex(_str)

    @abstractmethod
    def _get_validator_regexp(self, _str, max_value):
        pass


class ColonFormatter(Formatter):
    def format(self, value, max_value):
        return per_join(self._to_clean_str(value, max_value), ':', 2)

    def _get_validator_regexp(self, _str, max_value):
        return self._common_regex(max_value, r'\:', 2)


class PeriodFormatter(Formatter):
    def __init__(self, step=2):
        super().__init__()
        self.step = step

    def format(self, value, max_value):
        return per_join(self._to_clean_str(value, max_value), '.', self.step)

    def _get_validator_regexp(self, _str, max_value):
        return self._common_regex(max_value, r'\.', self.step)


class HyphenFormatter(Formatter):
    def format(self, value, max_value):
        return per_join(self._to_clean_str(value, max_value), '-', 2)

    def _get_validator_regexp(self, _str, max_value):
        return self._common_regex(max_value, r'\-', 2)


class CleanFormatter(Formatter):
    def format(self, value, max_value):
        return self._to_clean_str(value, max_value)

    def _get_validator_regexp(self, _str, max_value):
        return self._common_regex(max_value, '', 2)


DEFAULT_FORMATTERS = [
    ColonFormatter(),
    PeriodFormatter(2),
    PeriodFormatter(3),
    PeriodFormatter(4),
    HyphenFormatter(),
    CleanFormatter()
]



from .formatter import DEFAULT_FORMATTERS
from .utils import clamp, parse_hex


def mac_address(value=0, formatter=DEFAULT_FORMATTERS[0]):
    return MacAddress(value, formatter)


MAX_HEXADECIMAL_DIGITS = 12


class MacAddress:
    def __init__(self, value, formatter):
        self.formatter = formatter
        self.value = self._normalize_value(value)

    def _normalize_value(self, value):
        if isinstance(value, str):
            value_int = self._parse(value)
        elif isinstance(value, MacAddress):
            value_int = value.value
        elif isinstance(value, int):
            value_int = value
        else:
            raise ValueError('Invalid value to create MAC address')

        return clamp(value_int, self.max_value)

    @property
    def max_value(self):
        return pow(0x10, MAX_HEXADECIMAL_DIGITS) - 1

    def to_str(self, formatter=None):
        if formatter is None:
            formatter = self.formatter

        return formatter.format(self.value, self.max_value)

    def _parse(self, value_str):
        for each in [self.formatter] + DEFAULT_FORMATTERS:
            try:
                return each.parse(value_str, self.max_value)
            except ValueError:
                pass

        raise ValueError('Invalid MAC address format')

    def __str__(self):
        return self.to_str()

    def __int__(self):
        return self.value

    def __add__(self, val):
        return MacAddress(self.value + int(MacAddress(val, self.formatter)), self.formatter)

    def __radd__(self, val):
        return self + val

    def __sub__(self, val):
        return self + int(MacAddress(val, self.formatter))*(-1)

    def __eq__(self, rhs):
        return isinstance(rhs, MacAddress) and self.value == rhs.value

    def __hash__(self):
        return hash(self.value)



from os import path

from setuptools import setup

current_dir = path.abspath(path.dirname(__file__))
with open(path.join(current_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

version = '1.0.1'

setup(
    name='l2address',
    packages=['l2address'],
    version=version,
    license='MIT',
    description='MAC address manipulation library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Egor Blagov',
    author_email='e.m.blagov@gmail.com',
    url='https://github.com/EgorBlagov/l2address',
    download_url='https://github.com/EgorBlagov/l2address/archive/v{}.tar.gz'.format(
        version),
    keywords=['Networking', 'Telecommunication', 'MAC', 'L2'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



import math
import re


def clamp(value, max_limit):
    return value % (max_limit+1)


def parse_hex(value):
    return int(''.join([x.group(0) for x in re.finditer(r'[\da-fA-F]', value.lower())]), base=16)


def per_join(iterable, delimeter, step=1):
    return delimeter.join([''.join(iterable[step*i:step*i+step]) for i in range(math.ceil(len(iterable)/step))])
